import pytest

import aesara
import numpy as np
import sys, os

import gempy_legacy.core.data_modules.stack

sys.path.append("../..")
import gempy_legacy as gp


input_path = os.path.dirname(__file__) + '/../../examples/data'






