import pytest

import aesara
import numpy as np
import sys, os

sys.path.append("../..")
import gempy_legacy
import matplotlib.pyplot as plt
import pdb


@pytest.mark.skip()
class TestPerformance:
    @pytest.mark.skip()
    def test_high_res(self):
        pass
